import {computed, reactive, ref, toRef, toRefs, watch, watchEffect} from 'vue';
import type { App, Ref } from 'vue';
import axios from 'axios';
import type {AxiosInstance, AxiosRequestConfig, AxiosResponse } from 'axios';

export interface FormData {
    [key: string]: any;
}

export interface ValidationSchema {
    [key: string]: 'string' | 'number' | 'boolean' | 'array';
}

export interface FormState {
    initialData: FormData;
    data: FormData;
    errors: Record<string, string>;
    processing: boolean;
    //is_dirty: Ref<boolean>;
    [key: string]: any;
}

type SubmitMethod = 'POST' | 'GET' | 'PUT' | 'DELETE' | 'PATCH';

let sharedAxiosInstance: AxiosInstance | null = null;
let sharedFetchInstance: typeof fetch | null = null;

export function useForm(initialData: FormData, validationSchema: ValidationSchema = {}, axiosInstance?: AxiosInstance, fetchInstance?: typeof fetch) {
    const initialDataCopy = JSON.parse(JSON.stringify(initialData));
    const is_dirty = ref(false);

    const form = reactive<FormState>({
        initialData: { ...initialData },
        data: { ...initialData },
        errors: {},
        processing: false,
    });

    const formData = computed(() => form.data);

    function setErrors(errors: Record<string, string>) {
        form.errors = errors;
    }

    function clearErrors() {
        form.errors = {};
    }

    function startProcessing() {
        form.processing = true;
    }

    function stopProcessing() {
        form.processing = false;
    }

    function validate(): boolean {
        const errors: Record<string, string> = {};
        for (const key in validationSchema) {
            const value = form[key];
            const type = validationSchema[key];
            if (type === 'string' && typeof value !== 'string') {
                errors[key] = `${key} must be a string`;
            } else if (type === 'number' && typeof value !== 'number') {
                errors[key] = `${key} must be a number`;
            } else if (type === 'boolean' && typeof value !== 'boolean') {
                errors[key] = `${key} must be a boolean`;
            } else if (type === 'array' && !Array.isArray(value)) {
                errors[key] = `${key} must be an array`;
            }
        }
        setErrors(errors);
        return Object.keys(errors).length === 0;
    }

    async function submit(method: SubmitMethod, url: string, options: AxiosRequestConfig = {}): Promise<AxiosResponse> {
        if (!validate()) {
            throw new Error('Validation failed');
        }

        startProcessing();
        clearErrors();

        let axiosOrFetchInstance = axiosInstance || sharedAxiosInstance || axios.create();

        try {
            return await axiosOrFetchInstance({
                method,
                url,
                data: { ...form },
                ...options,
            });

        } catch (error) {
            if (axios.isAxiosError(error) && error.response) {
                setErrors(error.response.data.errors || {});
                throw new Error(error.response.data.message || 'Form submission failed');
            }
            throw error;
        } finally {
            stopProcessing();
        }
    }

    function checkIfDirty() {
        const initialDataString = JSON.stringify(initialDataCopy);
        const currentDataString = JSON.stringify(form.data);
        is_dirty.value = initialDataString !== currentDataString;
    }

    watch(() => form.data, () => {
        checkIfDirty();
    }, { deep: true });

    function reset() {
        Object.assign(form.data, initialDataCopy);
        is_dirty.value = false;
        clearErrors();
    }


    return reactive({
        ...form,
        submit,
        is_dirty,
        clearErrors,
        validate,
        reset,
    });

}

export function createFormPlugin(axiosInstance?: AxiosInstance, fetchInstance?: typeof fetch) {

    sharedAxiosInstance = axiosInstance || null;
    sharedFetchInstance = fetchInstance || null;


    return {
        install(app: App) {
            // This is mainly used in Vue 3 Options API
            app.config.globalProperties.$useForm = useForm;

            // This is used in Vue 3 Composition API (inject)
            app.provide('$useForm', useForm);
        },
    };
}
